# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES), and is copyright (c) 2018-2021
# by the software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia University
# Research Corporation, et al.  All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and
# license information.
#################################################################################
import pyomo.common.unittest as unittest

import pyomo.environ as pyo
from pyomo.common.collections import ComponentSet
from pyomo.core.expr.visitor import identify_variables
from pyomo.core.expr.compare import compare_expressions
from pyomo.contrib.mpc.modeling.cost_expressions import (
    get_penalty_from_constant_target,
    get_penalty_from_piecewise_constant_target,
    get_penalty_from_time_varying_target,
    get_penalty_from_target,
)
from pyomo.contrib.mpc.data.scalar_data import ScalarData
from pyomo.contrib.mpc.data.series_data import TimeSeriesData
from pyomo.contrib.mpc.data.interval_data import IntervalData


class TestTrackingCostConstantSetpoint(unittest.TestCase):
    def test_penalty_no_weights(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = ScalarData({m.v1[:]: 3.0, m.v2[:]: 4.0})
        variables = [m.v1, m.v2]
        m.var_set, m.tracking_expr = get_penalty_from_constant_target(
            variables, m.time, setpoint_data
        )
        self.assertEqual(len(m.var_set), 2)
        self.assertIn(0, m.var_set)
        self.assertIn(1, m.var_set)

        var_sets = {
            (i, t): ComponentSet(identify_variables(m.tracking_expr[i, t]))
            for i in m.var_set
            for t in m.time
        }
        for i in m.time:
            for j in m.var_set:
                self.assertIn(variables[j][i], var_sets[j, i])
                pred_value = (1 * i - 3) ** 2 if j == 0 else (2 * i - 4) ** 2
                self.assertEqual(pred_value, pyo.value(m.tracking_expr[j, i]))
                pred_expr = (m.v1[i] - 3) ** 2 if j == 0 else (m.v2[i] - 4) ** 2
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_expr[j, i].expr)
                )

    def test_penalty_with_weights(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = ScalarData({m.v1[:]: 3.0, m.v2[:]: 4.0})
        weight_data = ScalarData({m.v1[:]: 0.1, m.v2[:]: 0.5})
        m.var_set = pyo.Set(initialize=[0, 1])
        variables = [m.v1, m.v2]
        new_set, m.tracking_expr = get_penalty_from_constant_target(
            variables,
            m.time,
            setpoint_data,
            weight_data=weight_data,
            variable_set=m.var_set,
        )
        self.assertIs(new_set, m.var_set)

        var_sets = {
            (i, t): ComponentSet(identify_variables(m.tracking_expr[i, t]))
            for i in m.var_set
            for t in m.time
        }
        for i in m.time:
            for j in m.var_set:
                self.assertIn(variables[j][i], var_sets[j, i])
                pred_value = (
                    0.1 * (1 * i - 3) ** 2 if j == 0 else 0.5 * (2 * i - 4) ** 2
                )
                self.assertAlmostEqual(pred_value, pyo.value(m.tracking_expr[j, i]))
                pred_expr = (
                    0.1 * (m.v1[i] - 3) ** 2 if j == 0 else 0.5 * (m.v2[i] - 4) ** 2
                )
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_expr[j, i].expr)
                )

    def test_exceptions(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = ScalarData({m.v1[:]: 3.0})
        weight_data = ScalarData({m.v2[:]: 0.1})
        with self.assertRaisesRegex(KeyError, "Setpoint data"):
            _, m.tracking_expr = get_penalty_from_constant_target(
                [m.v1, m.v2], m.time, setpoint_data
            )

        setpoint_data = ScalarData({m.v1[:]: 3.0, m.v2[:]: 4.0})
        with self.assertRaisesRegex(KeyError, "Tracking weight"):
            _, m.tracking_expr = get_penalty_from_constant_target(
                [m.v1, m.v2], m.time, setpoint_data, weight_data=weight_data
            )

    def test_add_set_after_expr(self):
        # A small gotcha that may come up. This is known behavior
        # due to Pyomo's "implicit set" addition.
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[1, 2, 3])
        m.v1 = pyo.Var(m.time, initialize={i: 1 * i for i in m.time})
        m.v2 = pyo.Var(m.time, initialize={i: 2 * i for i in m.time})

        setpoint_data = ScalarData({m.v1[:]: 3.0, m.v2[:]: 4.0})
        weight_data = ScalarData({m.v1[:]: 0.1, m.v2[:]: 0.5})
        m.var_set = pyo.Set(initialize=[0, 1])
        variables = [m.v1, m.v2]
        new_set, tr_expr = get_penalty_from_constant_target(
            variables,
            m.time,
            setpoint_data,
            weight_data=weight_data,
            variable_set=m.var_set,
        )
        m.tracking_expr = tr_expr  # new_set gets added and assigned a name
        msg = "Attempting to re-assign"
        with self.assertRaisesRegex(RuntimeError, msg):
            # attempting to add the same component twice
            m.variable_set = new_set


class TestTrackingCostPiecewiseSetpoint(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        return m

    def test_piecewise_penalty_no_weights(self):
        m = self._make_model(n_time_points=5)

        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        setpoint_data = IntervalData(
            {m.var[:, "A"]: [2.0, 2.5], m.var[:, "B"]: [3.0, 3.5]}, [(0, 2), (2, 4)]
        )
        m.var_set, m.tracking_cost = get_penalty_from_piecewise_constant_target(
            variables, m.time, setpoint_data
        )
        for i in m.time:
            for j in m.var_set:
                if i <= 2:
                    pred_expr = (
                        (m.var[i, "A"] - 2.0) ** 2
                        if j == 0
                        else (m.var[i, "B"] - 3.0) ** 2
                    )
                else:
                    pred_expr = (
                        (m.var[i, "A"] - 2.5) ** 2
                        if j == 0
                        else (m.var[i, "B"] - 3.5) ** 2
                    )
                pred_value = pyo.value(pred_expr)
                self.assertEqual(pred_value, pyo.value(m.tracking_cost[j, i]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[j, i].expr)
                )

    def test_piecewise_penalty_with_weights(self):
        m = self._make_model(n_time_points=5)

        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        setpoint_data = IntervalData(
            {m.var[:, "A"]: [2.0, 2.5], m.var[:, "B"]: [3.0, 3.5]}, [(0, 2), (2, 4)]
        )
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.var[:, "B"]): 0.1,
        }
        m.var_set, m.tracking_cost = get_penalty_from_piecewise_constant_target(
            variables, m.time, setpoint_data, weight_data=weight_data
        )
        for i in m.time:
            for j in m.var_set:
                if i <= 2:
                    pred_expr = (
                        10.0 * (m.var[i, "A"] - 2.0) ** 2
                        if j == 0
                        else 0.1 * (m.var[i, "B"] - 3.0) ** 2
                    )
                else:
                    pred_expr = (
                        10.0 * (m.var[i, "A"] - 2.5) ** 2
                        if j == 0
                        else 0.1 * (m.var[i, "B"] - 3.5) ** 2
                    )
                pred_value = pyo.value(pred_expr)
                self.assertEqual(pred_value, pyo.value(m.tracking_cost[j, i]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[j, i].expr)
                )

    def test_piecewise_penalty_exceptions(self):
        m = self._make_model(n_time_points=5)

        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        setpoint_data = IntervalData({m.var[:, "A"]: [2.0, 2.5]}, [(0, 2), (2, 4)])
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.var[:, "B"]): 0.1,
        }
        msg = "Setpoint data does not contain"
        with self.assertRaisesRegex(KeyError, msg):
            tr_cost = get_penalty_from_piecewise_constant_target(
                variables, m.time, setpoint_data, weight_data=weight_data
            )

        setpoint_data = IntervalData(
            {m.var[:, "A"]: [2.0, 2.5], m.var[:, "B"]: [3.0, 3.5]}, [(0, 2), (2, 4)]
        )
        weight_data = {pyo.ComponentUID(m.var[:, "A"]): 10.0}
        msg = "Tracking weight does not contain"
        with self.assertRaisesRegex(KeyError, msg):
            tr_cost = get_penalty_from_piecewise_constant_target(
                variables, m.time, setpoint_data, weight_data=weight_data
            )


class TestTrackingCostVaryingSetpoint(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        return m

    def test_varying_setpoint_no_weights(self):
        m = self._make_model(n_time_points=5)
        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        A_setpoint = [1.0 - 0.1 * i for i in range(len(m.time))]
        B_setpoint = [5.0 + 0.1 * i for i in range(len(m.time))]
        setpoint_data = TimeSeriesData(
            {m.var[:, "A"]: A_setpoint, m.var[:, "B"]: B_setpoint}, m.time
        )
        m.var_set, m.tracking_cost = get_penalty_from_time_varying_target(
            variables, m.time, setpoint_data
        )
        for i, t in enumerate(m.time):
            for j in m.var_set:
                pred_expr = (
                    (m.var[t, "A"] - A_setpoint[i]) ** 2
                    if j == 0
                    else (m.var[t, "B"] - B_setpoint[i]) ** 2
                )
                pred_value = pyo.value(pred_expr)
                self.assertEqual(pred_value, pyo.value(m.tracking_cost[j, t]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[j, t].expr)
                )

    def test_varying_setpoint_with_weights(self):
        m = self._make_model(n_time_points=5)
        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        A_setpoint = [1.0 - 0.1 * i for i in range(len(m.time))]
        B_setpoint = [5.0 + 0.1 * i for i in range(len(m.time))]
        setpoint_data = TimeSeriesData(
            {m.var[:, "A"]: A_setpoint, m.var[:, "B"]: B_setpoint}, m.time
        )
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.var[:, "B"]): 0.1,
        }
        m.var_set, m.tracking_cost = get_penalty_from_time_varying_target(
            variables, m.time, setpoint_data, weight_data=weight_data
        )
        for i, t in enumerate(m.time):
            for j in m.var_set:
                pred_expr = (
                    10.0 * (m.var[t, "A"] - A_setpoint[i]) ** 2
                    if j == 0
                    else 0.1 * (m.var[t, "B"] - B_setpoint[i]) ** 2
                )
                pred_value = pyo.value(pred_expr)
                self.assertEqual(pred_value, pyo.value(m.tracking_cost[j, t]))
                self.assertTrue(
                    compare_expressions(pred_expr, m.tracking_cost[j, t].expr)
                )

    def test_varying_setpoint_exceptions(self):
        m = self._make_model(n_time_points=5)
        variables = [pyo.Reference(m.var[:, "A"]), pyo.Reference(m.var[:, "B"])]
        A_setpoint = [1.0 - 0.1 * i for i in range(len(m.time))]
        B_setpoint = [5.0 + 0.1 * i for i in range(len(m.time))]
        setpoint_data = TimeSeriesData(
            {m.var[:, "A"]: A_setpoint, m.var[:, "B"]: B_setpoint},
            [i + 10 for i in m.time],
        )
        weight_data = {
            pyo.ComponentUID(m.var[:, "A"]): 10.0,
            pyo.ComponentUID(m.var[:, "B"]): 0.1,
        }
        msg = "Mismatch in time points"
        with self.assertRaisesRegex(RuntimeError, msg):
            # Time-varying setpoint specifies different time points
            # from our time set.
            var_set, tr_cost = get_penalty_from_time_varying_target(
                variables, m.time, setpoint_data, weight_data=weight_data
            )

        setpoint_data = TimeSeriesData({m.var[:, "A"]: A_setpoint}, m.time)
        msg = "Setpoint data does not contain"
        with self.assertRaisesRegex(KeyError, msg):
            var_set, tr_cost = get_penalty_from_time_varying_target(
                variables, m.time, setpoint_data, weight_data=weight_data
            )

        setpoint_data = TimeSeriesData(
            {m.var[:, "A"]: A_setpoint, m.var[:, "B"]: B_setpoint}, m.time
        )
        weight_data = {pyo.ComponentUID(m.var[:, "A"]): 10.0}
        msg = "Tracking weight does not contain"
        with self.assertRaisesRegex(KeyError, msg):
            tr_cost = get_penalty_from_time_varying_target(
                variables, m.time, setpoint_data, weight_data=weight_data
            )


class TestGetPenaltyFromTarget(unittest.TestCase):
    def _make_model(self, n_time_points=3):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=list(range(n_time_points)))
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(
            m.time, m.comp, initialize={(i, j): 1.1 * i for i, j in m.time * m.comp}
        )
        return m

    def test_constant_setpoint(self):
        m = self._make_model()
        setpoint = {m.var[:, "A"]: 0.3, m.var[:, "B"]: 0.4}
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)
        pred_expr = {
            (i, t): (
                (m.var[t, "B"] - 0.4) ** 2 if i == 0 else (m.var[t, "A"] - 0.3) ** 2
            )
            for i, t in m.var_set * m.time
        }
        for t in m.time:
            for i in m.var_set:
                self.assertTrue(
                    compare_expressions(pred_expr[i, t], m.penalty[i, t].expr)
                )
                self.assertEqual(pyo.value(pred_expr[i, t]), pyo.value(m.penalty[i, t]))

    def test_constant_setpoint_with_ScalarData(self):
        m = self._make_model()
        setpoint = ScalarData({m.var[:, "A"]: 0.3, m.var[:, "B"]: 0.4})
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)
        pred_expr = {
            (i, t): (
                (m.var[t, "B"] - 0.4) ** 2 if i == 0 else (m.var[t, "A"] - 0.3) ** 2
            )
            for i, t in m.var_set * m.time
        }
        for t in m.time:
            for i in m.var_set:
                self.assertTrue(
                    compare_expressions(pred_expr[i, t], m.penalty[i, t].expr)
                )
                self.assertEqual(pyo.value(pred_expr[i, t]), pyo.value(m.penalty[i, t]))

    def test_varying_setpoint(self):
        m = self._make_model(n_time_points=5)
        A_target = [0.4, 0.6, 0.1, 0.0, 1.1]
        B_target = [0.8, 0.9, 1.3, 1.5, 1.4]
        setpoint = ({m.var[:, "A"]: A_target, m.var[:, "B"]: B_target}, m.time)
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)

        target = {
            (i, t): A_target[j] if i == 1 else B_target[t]
            for i in m.var_set
            for (j, t) in enumerate(m.time)
        }
        for i, t in m.var_set * m.time:
            pred_expr = (variables[i][t] - target[i, t]) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))

    def test_piecewise_constant_setpoint(self):
        m = self._make_model(n_time_points=5)
        A_target = [0.3, 0.9, 0.7]
        B_target = [1.1, 0.1, 0.5]
        setpoint = (
            {m.var[:, "A"]: A_target, m.var[:, "B"]: B_target},
            [(0.0, 0.0), (0.0, 2.0), (2.0, 4.0)],
        )
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)
        target = {
            (i, j): A_target[j] if i == 1 else B_target[j]
            for i in m.var_set
            for j in range(len(A_target))
        }
        for i, t in m.var_set * m.time:
            if t == 0:
                idx = 0
            elif t <= 2.0:
                idx = 1
            elif t <= 4.0:
                idx = 2
            pred_expr = (variables[i][t] - target[i, idx]) ** 2
            self.assertTrue(compare_expressions(pred_expr, m.penalty[i, t].expr))
            self.assertEqual(pyo.value(pred_expr), pyo.value(m.penalty[i, t]))

    def test_bad_argument(self):
        m = self._make_model(n_time_points=3)
        A_target = [0.4, 0.6, 0.1]
        B_target = [0.8, 0.9, 1.3]
        setpoint = ({m.var[:, "A"]: A_target, m.var[:, "B"]: B_target}, m.time)
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        msg = "tolerance.*can only be used"
        with self.assertRaisesRegex(RuntimeError, msg):
            m.var_set, m.penalty = get_penalty_from_target(
                variables, m.time, setpoint, tolerance=1e-8
            )

    def test_bad_data_tuple(self):
        m = self._make_model(n_time_points=3)
        A_target = [0.4, 0.6, 0.1]
        B_target = [0.8, 0.9, 1.3]
        setpoint = (
            {m.var[:, "A"]: A_target, m.var[:, "B"]: B_target},
            m.time,
            "something else",
        )
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        msg = "tuple of length two"
        with self.assertRaisesRegex(TypeError, msg):
            m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)

    def test_bad_data_tuple_entry_0(self):
        m = self._make_model(n_time_points=3)
        A_target = [0.4, 0.6, 0.1]
        B_target = [0.8, 0.9, 1.3]
        setpoint = ([(m.var[:, "A"], A_target), (m.var[:, "B"], B_target)], m.time)
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        msg = "must be instance of MutableMapping"
        with self.assertRaisesRegex(TypeError, msg):
            m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)

    def test_empty_time_list(self):
        m = self._make_model(n_time_points=3)
        A_target = []
        B_target = []
        setpoint = ({m.var[:, "A"]: A_target, m.var[:, "B"]: B_target}, [])
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        msg = "Time sequence.*is empty"
        with self.assertRaisesRegex(ValueError, msg):
            m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)

    def test_bad_time_list(self):
        m = self._make_model(n_time_points=3)
        A_target = [0.4, 0.6, 0.1]
        B_target = [0.8, 0.9, 1.3]
        setpoint = (
            dict([(m.var[:, "A"], A_target), (m.var[:, "B"], B_target)]),
            [0.0, (0.1, 0.2), 0.3],
        )
        variables = [pyo.Reference(m.var[:, "B"]), pyo.Reference(m.var[:, "A"])]
        msg = "Second entry of data tuple must be"
        with self.assertRaisesRegex(TypeError, msg):
            m.var_set, m.penalty = get_penalty_from_target(variables, m.time, setpoint)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import glob
import logging
import os
import shutil
import sys
import subprocess

from pyomo.common.download import FileDownloader
from pyomo.common.envvar import PYOMO_CONFIG_DIR
from pyomo.common.fileutils import find_library, this_file_dir
from pyomo.common.tempfiles import TempfileManager

logger = logging.getLogger(__name__ if __name__ != '__main__' else 'pyomo')


def build_ginac_library(parallel=None, argv=None, env=None):
    sys.stdout.write("\n**** Building GiNaC library ****\n")

    configure_cmd = [
        os.path.join('.', 'configure'),
        '--prefix=' + PYOMO_CONFIG_DIR,
        '--disable-static',
    ]
    make_cmd = ['make']
    if parallel:
        make_cmd.append(f'-j{parallel}')
    install_cmd = ['make', 'install']

    env = dict(os.environ)
    pcdir = os.path.join(PYOMO_CONFIG_DIR, 'lib', 'pkgconfig')
    if 'PKG_CONFIG_PATH' in env:
        pcdir += os.pathsep + env['PKG_CONFIG_PATH']
    env['PKG_CONFIG_PATH'] = pcdir

    with TempfileManager.new_context() as tempfile:
        tmpdir = tempfile.mkdtemp()

        downloader = FileDownloader()
        if argv:
            downloader.parse_args(argv)

        url = 'https://www.ginac.de/CLN/cln-1.3.7.tar.bz2'
        cln_dir = os.path.join(tmpdir, 'cln')
        downloader.set_destination_filename(cln_dir)
        logger.info(
            "Fetching CLN from %s and installing it to %s"
            % (url, downloader.destination())
        )
        downloader.get_tar_archive(url, dirOffset=1)
        assert subprocess.run(configure_cmd, cwd=cln_dir, env=env).returncode == 0
        logger.info("\nBuilding CLN\n")
        assert subprocess.run(make_cmd, cwd=cln_dir, env=env).returncode == 0
        assert subprocess.run(install_cmd, cwd=cln_dir, env=env).returncode == 0

        url = 'https://www.ginac.de/ginac-1.8.10.tar.bz2'
        ginac_dir = os.path.join(tmpdir, 'ginac')
        downloader.set_destination_filename(ginac_dir)
        logger.info(
            "Fetching GiNaC from %s and installing it to %s"
            % (url, downloader.destination())
        )
        downloader.get_tar_archive(url, dirOffset=1)
        assert subprocess.run(configure_cmd, cwd=ginac_dir, env=env).returncode == 0
        logger.info("\nBuilding GiNaC\n")
        assert subprocess.run(make_cmd, cwd=ginac_dir, env=env).returncode == 0
        assert subprocess.run(install_cmd, cwd=ginac_dir, env=env).returncode == 0
        print("Installed GiNaC to %s" % (ginac_dir,))


def _find_include(libdir, incpaths):
    rel_path = ('include',) + incpaths
    while 1:
        basedir = os.path.dirname(libdir)
        if not basedir or basedir == libdir:
            return None
        if os.path.exists(os.path.join(basedir, *rel_path)):
            return os.path.join(basedir, *(rel_path[: -len(incpaths)]))
        libdir = basedir


def build_ginac_interface(parallel=None, args=None):
    from distutils.dist import Distribution
    from pybind11.setup_helpers import Pybind11Extension, build_ext
    from pyomo.common.cmake_builder import handleReadonly

    sys.stdout.write("\n**** Building GiNaC interface ****\n")

    if args is None:
        args = []
    sources = [
        os.path.join(this_file_dir(), 'ginac', 'src', fname)
        for fname in ['ginac_interface.cpp']
    ]

    ginac_lib = find_library('ginac')
    if not ginac_lib:
        raise RuntimeError(
            'could not find the GiNaC library; please make sure either to install '
            'the library and development headers system-wide, or include the '
            'path to the library in the LD_LIBRARY_PATH environment variable'
        )
    ginac_lib_dir = os.path.dirname(ginac_lib)
    ginac_include_dir = _find_include(ginac_lib_dir, ('ginac', 'ginac.h'))
    if not ginac_include_dir:
        raise RuntimeError('could not find GiNaC include directory')

    cln_lib = find_library('cln')
    if not cln_lib:
        raise RuntimeError(
            'could not find the CLN library; please make sure either to install '
            'the library and development headers system-wide, or include the '
            'path to the library in the LD_LIBRARY_PATH environment variable'
        )
    cln_lib_dir = os.path.dirname(cln_lib)
    cln_include_dir = _find_include(cln_lib_dir, ('cln', 'cln.h'))
    if not cln_include_dir:
        raise RuntimeError('could not find CLN include directory')

    extra_args = ['-std=c++11']
    ext = Pybind11Extension(
        'ginac_interface',
        sources=sources,
        language='c++',
        include_dirs=[cln_include_dir, ginac_include_dir],
        library_dirs=[cln_lib_dir, ginac_lib_dir],
        libraries=['cln', 'ginac'],
        extra_compile_args=extra_args,
    )

    class ginacBuildExt(build_ext):
        def run(self):
            basedir = os.path.abspath(os.path.curdir)
            with TempfileManager.new_context() as tempfile:
                if self.inplace:
                    tmpdir = os.path.join(this_file_dir(), 'ginac')
                else:
                    tmpdir = os.path.abspath(tempfile.mkdtemp())
                sys.stdout.write("Building in '%s'\n" % tmpdir)
                os.chdir(tmpdir)
                super(ginacBuildExt, self).run()
                if not self.inplace:
                    library = glob.glob("build/*/ginac_interface.*")[0]
                    target = os.path.join(
                        PYOMO_CONFIG_DIR,
                        'lib',
                        'python%s.%s' % sys.version_info[:2],
                        'site-packages',
                        '.',
                    )
                    if not os.path.exists(target):
                        os.makedirs(target)
                    sys.stdout.write(f"Installing {library} in {target}\n")
                    shutil.copy(library, target)

    package_config = {
        'name': 'ginac_interface',
        'packages': [],
        'ext_modules': [ext],
        'cmdclass': {"build_ext": ginacBuildExt},
    }

    dist = Distribution(package_config)
    dist.script_args = ['build_ext'] + args
    dist.parse_command_line()
    dist.run_command('build_ext')


class GiNaCInterfaceBuilder:
    def __call__(self, parallel):
        return build_ginac_interface(parallel)

    def skip(self):
        return not find_library('ginac')


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-j",
        dest='parallel',
        type=int,
        default=None,
        help="Enable parallel build with PARALLEL cores",
    )
    parser.add_argument(
        "--build-deps",
        dest='build_deps',
        action='store_true',
        default=False,
        help="Download and build the CLN/GiNaC libraries",
    )
    options, argv = parser.parse_known_args(sys.argv)
    logging.getLogger('pyomo').setLevel(logging.INFO)
    if options.build_deps:
        build_ginac_library(options.parallel, [])
    build_ginac_interface(options.parallel, argv[1:])



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from ..dependencies import numpy_available, scipy_available

if numpy_available and scipy_available:
    from .block_vector import BlockVector, NotFullyDefinedBlockVectorError
    from .block_matrix import BlockMatrix, NotFullyDefinedBlockMatrixError



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import (
    pandas as pd,
    pandas_available,
    numpy as np,
    numpy_available,
)

import os.path
import json

import pyomo.environ as pyo

from pyomo.common.fileutils import this_file_dir
import pyomo.common.unittest as unittest

import pyomo.contrib.parmest.parmest as parmest
from pyomo.opt import SolverFactory

from pyomo.contrib.parmest.utils.model_utils import update_model_from_suffix
from pyomo.contrib.doe.examples.reactor_example import (
    ReactorExperiment as FullReactorExperiment,
)

currdir = this_file_dir()
file_path = os.path.join(currdir, "..", "..", "doe", "examples", "result.json")

with open(file_path) as f:
    data_ex = json.load(f)
data_ex["control_points"] = {float(k): v for k, v in data_ex["control_points"].items()}

ipopt_available = pyo.SolverFactory("ipopt").available()


@unittest.skipIf(
    not parmest.parmest_available,
    "Cannot test parmest: required dependencies are missing",
)
@unittest.skipIf(not ipopt_available, "The 'ipopt' solver is not available")
class TestUtils(unittest.TestCase):

    def test_convert_param_to_var(self):
        # TODO: Check that this works for different structured models (indexed, blocks, etc)

        from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
            ReactorDesignExperiment,
        )

        data = pd.DataFrame(
            data=[
                [1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
            ],
            columns=["sv", "caf", "ca", "cb", "cc", "cd"],
        )

        # make model
        exp = ReactorDesignExperiment(data, 0)
        instance = exp.get_labeled_model()

        theta_names = ['k1', 'k2', 'k3']
        m_vars = parmest.utils.convert_params_to_vars(
            instance, theta_names, fix_vars=True
        )

        for v in theta_names:
            self.assertTrue(hasattr(m_vars, v))
            c = m_vars.find_component(v)
            self.assertIsInstance(c, pyo.Var)
            self.assertTrue(c.fixed)
            c_old = instance.find_component(v)
            self.assertEqual(pyo.value(c), pyo.value(c_old))
            self.assertTrue(c in m_vars.unknown_parameters)

    def test_update_model_from_suffix_experiment_outputs(self):
        from pyomo.contrib.parmest.examples.reactor_design.reactor_design import (
            ReactorDesignExperiment,
        )

        data = pd.DataFrame(
            data=[
                [1.05, 10000, 3458.4, 1060.8, 1683.9, 1898.5],
                [1.10, 10000, 3535.1, 1064.8, 1613.3, 1893.4],
                [1.15, 10000, 3609.1, 1067.8, 1547.5, 1887.8],
            ],
            columns=["sv", "caf", "ca", "cb", "cc", "cd"],
        )
        experiment = ReactorDesignExperiment(data, 0)
        test_model = experiment.get_labeled_model()

        suffix_obj = test_model.experiment_outputs  # a Suffix
        var_list = list(suffix_obj.keys())  # components
        orig_var_vals = np.array([pyo.value(v) for v in var_list])
        orig_suffix_val = np.array([tag for _, tag in suffix_obj.items()])
        new_vals = orig_var_vals + 0.5
        # Update the model from the suffix
        update_model_from_suffix(suffix_obj, new_vals)
        # ── Check results ────────────────────────────────────────────────────
        new_var_vals = np.array([pyo.value(v) for v in var_list])
        new_suffix_val = np.array(list(suffix_obj.values()))
        # (1) Variables have been overwritten with `new_vals`
        self.assertTrue(np.allclose(new_var_vals, new_vals))
        # (2) Suffix tags are unchanged
        self.assertTrue(np.array_equal(new_suffix_val, orig_suffix_val))

    def test_update_model_from_suffix_measurement_error(self):
        experiment = FullReactorExperiment(data_ex, 10, 3)
        test_model = experiment.get_labeled_model()

        suffix_obj = test_model.measurement_error  # a Suffix
        var_list = list(suffix_obj.keys())  # components
        orig_var_vals = np.array([suffix_obj[v] for v in var_list])
        new_vals = orig_var_vals + 0.5
        # Update the model from the suffix
        update_model_from_suffix(suffix_obj, new_vals)
        # ── Check results ────────────────────────────────────────────────────
        new_var_vals = np.array([suffix_obj[v] for v in var_list])
        # (1) Variables have been overwritten with `new_vals`
        self.assertTrue(np.allclose(new_var_vals, new_vals))

    def test_update_model_from_suffix_length_mismatch(self):
        m = pyo.ConcreteModel()

        # Create a suffix with a Var component
        m.x = pyo.Var(initialize=0.0)
        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters[m.x] = 0.0  # tag a Var
        with self.assertRaisesRegex(
            ValueError, "values length does not match suffix length"
        ):
            # Attempt to update with a list of different length
            update_model_from_suffix(m.unknown_parameters, [42, 43, 44])

    def test_update_model_from_suffix_not_numeric(self):
        m = pyo.ConcreteModel()

        # Create a suffix with a Var component
        m.x = pyo.Var(initialize=0.0)
        m.y = pyo.Var(initialize=1.0)
        bad_value = "not_a_number"
        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        # Make multiple values
        m.unknown_parameters[m.x] = 0.0  # tag a Var
        m.unknown_parameters[m.y] = bad_value  # tag a Var with a bad value
        # Attempt to update with a list of mixed types
        # This should raise an error because this utility only allows numeric values
        # in the model to be updated.

        with self.assertRaisesRegex(
            ValueError, f"could not convert string to float: '{bad_value}'"
        ):
            # Attempt to update with a non-numeric value
            update_model_from_suffix(m.unknown_parameters, [42, bad_value])

    def test_update_model_from_suffix_wrong_component_type(self):
        m = pyo.ConcreteModel()

        # Create a suffix with a Var component
        m.x = pyo.Var(initialize=0.0)
        m.e = pyo.Expression(expr=m.x + 1)  # not Var/Param
        m.unknown_parameters = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.unknown_parameters[m.x] = 0.0
        m.unknown_parameters[m.e] = 1.0  # tag an Expression
        # Attempt to update with a list of wrong component type
        with self.assertRaisesRegex(
            TypeError,
            f"Unsupported component type {type(m.e)}; expected VarData or ParamData.",
        ):
            update_model_from_suffix(m.unknown_parameters, [42, 43])

    def test_update_model_from_suffix_unsupported_component(self):
        m = pyo.ConcreteModel()

        # Create a suffix with a ConstraintData component
        m.x = pyo.Var(initialize=0.0)
        m.c = pyo.Constraint(expr=m.x == 0)  # not Var/Param!

        m.bad_suffix = pyo.Suffix(direction=pyo.Suffix.LOCAL)
        m.bad_suffix[m.c] = 0  # tag a Constraint

        with self.assertRaisesRegex(
            TypeError, r"Unsupported component type .*Constraint.*"
        ):
            update_model_from_suffix(m.bad_suffix, [1.0])

    def test_update_model_from_suffix_empty(self):
        m = pyo.ConcreteModel()

        # Create an empty suffix
        m.empty_suffix = pyo.Suffix(direction=pyo.Suffix.LOCAL)

        # This should not raise an error
        update_model_from_suffix(m.empty_suffix, [])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.contrib.pynumero.dependencies import numpy as np, scipy


def check_vectors_specific_order(tst, v1, v1order, v2, v2order, v1_v2_map=None):
    tst.assertEqual(len(v1), len(v1order))
    tst.assertEqual(len(v2), len(v2order))
    tst.assertEqual(len(v1), len(v2))
    if v1_v2_map is None:
        v2map = {s: v2order.index(s) for s in v1order}
    else:
        v2map = {s: v2order.index(v1_v2_map[s]) for s in v1order}
    for i, s in enumerate(v1order):
        tst.assertAlmostEqual(v1[i], v2[v2map[s]], places=7)


def check_sparse_matrix_specific_order(
    tst,
    m1,
    m1rows,
    m1cols,
    m2,
    m2rows,
    m2cols,
    m1_m2_rows_map=None,
    m1_m2_cols_map=None,
):
    tst.assertEqual(m1.shape[0], len(m1rows))
    tst.assertEqual(m1.shape[1], len(m1cols))
    tst.assertEqual(m2.shape[0], len(m2rows))
    tst.assertEqual(m2.shape[1], len(m2cols))
    tst.assertEqual(len(m1rows), len(m2rows))
    tst.assertEqual(len(m1cols), len(m2cols))

    m1c = m1
    if scipy.sparse.issparse(m1c):
        m1c = m1c.todense()
    m2d = m2
    if scipy.sparse.issparse(m2d):
        m2d = m2d.todense()

    m2c = np.zeros((len(m2rows), len(m2cols)))
    if m1_m2_rows_map is None:
        rowmap = [m2rows.index(x) for x in m1rows]
    else:
        rowmap = [m2rows.index(m1_m2_rows_map[x]) for x in m1rows]
    if m1_m2_cols_map is None:
        colmap = [m2cols.index(x) for x in m1cols]
    else:
        colmap = [m2cols.index(m1_m2_cols_map[x]) for x in m1cols]

    for i in range(len(m1rows)):
        for j in range(len(m1cols)):
            m2c[i, j] = m2d[rowmap[i], colmap[j]]

    for i in range(len(m1rows)):
        for j in range(len(m1cols)):
            tst.assertAlmostEqual(m1c[i, j], m2c[i, j], places=7)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common import unittest
from pyomo.contrib.appsi.solvers import ipopt

ipopt_available = ipopt.Ipopt().available()


@unittest.skipIf(not ipopt_available, "The 'ipopt' command is not available")
class TestIpoptInterface(unittest.TestCase):
    def test_has_linear_solver(self):
        opt = ipopt.Ipopt()
        self.assertTrue(
            any(
                map(
                    opt.has_linear_solver,
                    [
                        'mumps',
                        'ma27',
                        'ma57',
                        'ma77',
                        'ma86',
                        'ma97',
                        'pardiso',
                        'pardisomkl',
                        'spral',
                        'wsmp',
                    ],
                )
            )
        )
        self.assertFalse(opt.has_linear_solver('bogus_linear_solver'))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo

from pyomo.common.dependencies import mpi4py_available, numpy_available
from pyomo.contrib.benders.benders_cuts import BendersCutGenerator

ipopt_available = pyo.SolverFactory('ipopt').available(exception_flag=False)

for mip_name in (
    'cplex_direct',
    'gurobi_direct',
    'gurobi',
    'cplex',
    'glpk',
    'cbc',
    'highs',
):
    mip_available = pyo.SolverFactory(mip_name).available(exception_flag=False)
    if mip_available:
        break


@unittest.pytest.mark.mpi
class MPITestBenders(unittest.TestCase):
    @unittest.skipIf(not mpi4py_available, 'mpi4py is not available.')
    @unittest.skipIf(not numpy_available, 'numpy is not available.')
    @unittest.skipIf(not mip_available, 'MIP solver is not available.')
    def test_farmer(self):
        class Farmer:
            def __init__(self):
                self.crops = ['WHEAT', 'CORN', 'SUGAR_BEETS']
                self.total_acreage = 500
                self.PriceQuota = {
                    'WHEAT': 100000.0,
                    'CORN': 100000.0,
                    'SUGAR_BEETS': 6000.0,
                }
                self.SubQuotaSellingPrice = {
                    'WHEAT': 170.0,
                    'CORN': 150.0,
                    'SUGAR_BEETS': 36.0,
                }
                self.SuperQuotaSellingPrice = {
                    'WHEAT': 0.0,
                    'CORN': 0.0,
                    'SUGAR_BEETS': 10.0,
                }
                self.CattleFeedRequirement = {
                    'WHEAT': 200.0,
                    'CORN': 240.0,
                    'SUGAR_BEETS': 0.0,
                }
                self.PurchasePrice = {
                    'WHEAT': 238.0,
                    'CORN': 210.0,
                    'SUGAR_BEETS': 100000.0,
                }
                self.PlantingCostPerAcre = {
                    'WHEAT': 150.0,
                    'CORN': 230.0,
                    'SUGAR_BEETS': 260.0,
                }
                self.scenarios = [
                    'BelowAverageScenario',
                    'AverageScenario',
                    'AboveAverageScenario',
                ]
                self.crop_yield = dict()
                self.crop_yield['BelowAverageScenario'] = {
                    'WHEAT': 2.0,
                    'CORN': 2.4,
                    'SUGAR_BEETS': 16.0,
                }
                self.crop_yield['AverageScenario'] = {
                    'WHEAT': 2.5,
                    'CORN': 3.0,
                    'SUGAR_BEETS': 20.0,
                }
                self.crop_yield['AboveAverageScenario'] = {
                    'WHEAT': 3.0,
                    'CORN': 3.6,
                    'SUGAR_BEETS': 24.0,
                }
                self.scenario_probabilities = dict()
                self.scenario_probabilities['BelowAverageScenario'] = 0.3333
                self.scenario_probabilities['AverageScenario'] = 0.3334
                self.scenario_probabilities['AboveAverageScenario'] = 0.3333

        def create_root(farmer):
            m = pyo.ConcreteModel()

            m.crops = pyo.Set(initialize=farmer.crops, ordered=True)
            m.scenarios = pyo.Set(initialize=farmer.scenarios, ordered=True)

            m.devoted_acreage = pyo.Var(m.crops, bounds=(0, farmer.total_acreage))
            m.eta = pyo.Var(m.scenarios)
            for s in m.scenarios:
                m.eta[s].setlb(-432000 * farmer.scenario_probabilities[s])

            m.total_acreage_con = pyo.Constraint(
                expr=sum(m.devoted_acreage.values()) <= farmer.total_acreage
            )

            m.obj = pyo.Objective(
                expr=sum(
                    farmer.PlantingCostPerAcre[crop] * m.devoted_acreage[crop]
                    for crop in m.crops
                )
                + sum(m.eta.values())
            )
            return m

        def create_subproblem(root, farmer, scenario):
            m = pyo.ConcreteModel()

            m.crops = pyo.Set(initialize=farmer.crops, ordered=True)

            m.devoted_acreage = pyo.Var(m.crops)
            m.QuantitySubQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
            m.QuantitySuperQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
            m.QuantityPurchased = pyo.Var(m.crops, bounds=(0.0, None))

            def EnforceCattleFeedRequirement_rule(m, i):
                return (
                    farmer.CattleFeedRequirement[i]
                    <= (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
                    + m.QuantityPurchased[i]
                    - m.QuantitySubQuotaSold[i]
                    - m.QuantitySuperQuotaSold[i]
                )

            m.EnforceCattleFeedRequirement = pyo.Constraint(
                m.crops, rule=EnforceCattleFeedRequirement_rule
            )

            def LimitAmountSold_rule(m, i):
                return (
                    m.QuantitySubQuotaSold[i]
                    + m.QuantitySuperQuotaSold[i]
                    - (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
                    <= 0.0
                )

            m.LimitAmountSold = pyo.Constraint(m.crops, rule=LimitAmountSold_rule)

            def EnforceQuotas_rule(m, i):
                return (0.0, m.QuantitySubQuotaSold[i], farmer.PriceQuota[i])

            m.EnforceQuotas = pyo.Constraint(m.crops, rule=EnforceQuotas_rule)

            obj_expr = sum(
                farmer.PurchasePrice[crop] * m.QuantityPurchased[crop]
                for crop in m.crops
            )
            obj_expr -= sum(
                farmer.SubQuotaSellingPrice[crop] * m.QuantitySubQuotaSold[crop]
                for crop in m.crops
            )
            obj_expr -= sum(
                farmer.SuperQuotaSellingPrice[crop] * m.QuantitySuperQuotaSold[crop]
                for crop in m.crops
            )
            m.obj = pyo.Objective(
                expr=farmer.scenario_probabilities[scenario] * obj_expr
            )

            complicating_vars_map = pyo.ComponentMap()
            for crop in m.crops:
                complicating_vars_map[root.devoted_acreage[crop]] = m.devoted_acreage[
                    crop
                ]

            return m, complicating_vars_map

        farmer = Farmer()
        m = create_root(farmer=farmer)
        root_vars = list(m.devoted_acreage.values())
        m.benders = BendersCutGenerator()
        m.benders.set_input(root_vars=root_vars, tol=1e-8)
        for s in farmer.scenarios:
            subproblem_fn_kwargs = dict()
            subproblem_fn_kwargs['root'] = m
            subproblem_fn_kwargs['farmer'] = farmer
            subproblem_fn_kwargs['scenario'] = s
            m.benders.add_subproblem(
                subproblem_fn=create_subproblem,
                subproblem_fn_kwargs=subproblem_fn_kwargs,
                root_eta=m.eta[s],
                subproblem_solver=mip_name,
            )
        opt = pyo.SolverFactory(mip_name)

        for i in range(30):
            res = opt.solve(m, tee=False)
            cuts_added = m.benders.generate_cut()
            if len(cuts_added) == 0:
                break

        self.assertAlmostEqual(m.devoted_acreage['CORN'].value, 80, 7)
        self.assertAlmostEqual(m.devoted_acreage['SUGAR_BEETS'].value, 250, 7)
        self.assertAlmostEqual(m.devoted_acreage['WHEAT'].value, 170, 7)

    @unittest.skipIf(not mpi4py_available, 'mpi4py is not available.')
    @unittest.skipIf(not numpy_available, 'numpy is not available.')
    @unittest.skipIf(not ipopt_available, 'ipopt is not available.')
    def test_grothey(self):
        def create_root():
            m = pyo.ConcreteModel()
            m.y = pyo.Var(bounds=(1, None))
            m.eta = pyo.Var(bounds=(-10, None))
            m.obj = pyo.Objective(expr=m.y**2 + m.eta)
            return m

        def create_subproblem(root):
            m = pyo.ConcreteModel()
            m.x1 = pyo.Var()
            m.x2 = pyo.Var()
            m.y = pyo.Var()
            m.obj = pyo.Objective(expr=-m.x2)
            m.c1 = pyo.Constraint(expr=(m.x1 - 1) ** 2 + m.x2**2 <= pyo.log(m.y))
            m.c2 = pyo.Constraint(expr=(m.x1 + 1) ** 2 + m.x2**2 <= pyo.log(m.y))

            complicating_vars_map = pyo.ComponentMap()
            complicating_vars_map[root.y] = m.y

            return m, complicating_vars_map

        m = create_root()
        root_vars = [m.y]
        m.benders = BendersCutGenerator()
        m.benders.set_input(root_vars=root_vars, tol=1e-8)
        m.benders.add_subproblem(
            subproblem_fn=create_subproblem,
            subproblem_fn_kwargs={'root': m},
            root_eta=m.eta,
            subproblem_solver='ipopt',
        )
        opt = pyo.SolverFactory('ipopt')

        for i in range(30):
            res = opt.solve(m, tee=False)
            cuts_added = m.benders.generate_cut()
            if len(cuts_added) == 0:
                break
        self.assertAlmostEqual(m.y.value, 2.721381, 4)
        self.assertAlmostEqual(m.eta.value, -0.0337568, 4)

    @unittest.skipIf(not mpi4py_available, 'mpi4py is not available.')
    @unittest.skipIf(not numpy_available, 'numpy is not available.')
    @unittest.skipIf(not mip_available, 'MIP solver is not available.')
    def test_four_scen_farmer(self):
        class FourScenFarmer:
            def __init__(self):
                self.crops = ['WHEAT', 'CORN', 'SUGAR_BEETS']
                self.total_acreage = 500
                self.PriceQuota = {
                    'WHEAT': 100000.0,
                    'CORN': 100000.0,
                    'SUGAR_BEETS': 6000.0,
                }
                self.SubQuotaSellingPrice = {
                    'WHEAT': 170.0,
                    'CORN': 150.0,
                    'SUGAR_BEETS': 36.0,
                }
                self.SuperQuotaSellingPrice = {
                    'WHEAT': 0.0,
                    'CORN': 0.0,
                    'SUGAR_BEETS': 10.0,
                }
                self.CattleFeedRequirement = {
                    'WHEAT': 200.0,
                    'CORN': 240.0,
                    'SUGAR_BEETS': 0.0,
                }
                self.PurchasePrice = {
                    'WHEAT': 238.0,
                    'CORN': 210.0,
                    'SUGAR_BEETS': 100000.0,
                }
                self.PlantingCostPerAcre = {
                    'WHEAT': 150.0,
                    'CORN': 230.0,
                    'SUGAR_BEETS': 260.0,
                }
                self.scenarios = [
                    'BelowAverageScenario',
                    'AverageScenario',
                    'AboveAverageScenario',
                    'Scenario4',
                ]
                self.crop_yield = dict()
                self.crop_yield['BelowAverageScenario'] = {
                    'WHEAT': 2.0,
                    'CORN': 2.4,
                    'SUGAR_BEETS': 16.0,
                }
                self.crop_yield['AverageScenario'] = {
                    'WHEAT': 2.5,
                    'CORN': 3.0,
                    'SUGAR_BEETS': 20.0,
                }
                self.crop_yield['AboveAverageScenario'] = {
                    'WHEAT': 3.0,
                    'CORN': 3.6,
                    'SUGAR_BEETS': 24.0,
                }
                self.crop_yield['Scenario4'] = {
                    'WHEAT': 2.0,
                    'CORN': 3.0,
                    'SUGAR_BEETS': 24.0,
                }
                self.scenario_probabilities = dict()
                self.scenario_probabilities['BelowAverageScenario'] = 0.25
                self.scenario_probabilities['AverageScenario'] = 0.25
                self.scenario_probabilities['AboveAverageScenario'] = 0.25
                self.scenario_probabilities['Scenario4'] = 0.25

        def create_root(farmer):
            m = pyo.ConcreteModel()

            m.crops = pyo.Set(initialize=farmer.crops, ordered=True)
            m.scenarios = pyo.Set(initialize=farmer.scenarios, ordered=True)

            m.devoted_acreage = pyo.Var(m.crops, bounds=(0, farmer.total_acreage))
            m.eta = pyo.Var(m.scenarios)
            for s in m.scenarios:
                m.eta[s].setlb(-432000 * farmer.scenario_probabilities[s])

            m.total_acreage_con = pyo.Constraint(
                expr=sum(m.devoted_acreage.values()) <= farmer.total_acreage
            )

            m.obj = pyo.Objective(
                expr=sum(
                    farmer.PlantingCostPerAcre[crop] * m.devoted_acreage[crop]
                    for crop in m.crops
                )
                + sum(m.eta.values())
            )
            return m

        def create_subproblem(root, farmer, scenario):
            m = pyo.ConcreteModel()

            m.crops = pyo.Set(initialize=farmer.crops, ordered=True)

            m.devoted_acreage = pyo.Var(m.crops)
            m.QuantitySubQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
            m.QuantitySuperQuotaSold = pyo.Var(m.crops, bounds=(0.0, None))
            m.QuantityPurchased = pyo.Var(m.crops, bounds=(0.0, None))

            def EnforceCattleFeedRequirement_rule(m, i):
                return (
                    farmer.CattleFeedRequirement[i]
                    <= (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
                    + m.QuantityPurchased[i]
                    - m.QuantitySubQuotaSold[i]
                    - m.QuantitySuperQuotaSold[i]
                )

            m.EnforceCattleFeedRequirement = pyo.Constraint(
                m.crops, rule=EnforceCattleFeedRequirement_rule
            )

            def LimitAmountSold_rule(m, i):
                return (
                    m.QuantitySubQuotaSold[i]
                    + m.QuantitySuperQuotaSold[i]
                    - (farmer.crop_yield[scenario][i] * m.devoted_acreage[i])
                    <= 0.0
                )

            m.LimitAmountSold = pyo.Constraint(m.crops, rule=LimitAmountSold_rule)

            def EnforceQuotas_rule(m, i):
                return (0.0, m.QuantitySubQuotaSold[i], farmer.PriceQuota[i])

            m.EnforceQuotas = pyo.Constraint(m.crops, rule=EnforceQuotas_rule)

            obj_expr = sum(
                farmer.PurchasePrice[crop] * m.QuantityPurchased[crop]
                for crop in m.crops
            )
            obj_expr -= sum(
                farmer.SubQuotaSellingPrice[crop] * m.QuantitySubQuotaSold[crop]
                for crop in m.crops
            )
            obj_expr -= sum(
                farmer.SuperQuotaSellingPrice[crop] * m.QuantitySuperQuotaSold[crop]
                for crop in m.crops
            )
            m.obj = pyo.Objective(
                expr=farmer.scenario_probabilities[scenario] * obj_expr
            )

            complicating_vars_map = pyo.ComponentMap()
            for crop in m.crops:
                complicating_vars_map[root.devoted_acreage[crop]] = m.devoted_acreage[
                    crop
                ]

            return m, complicating_vars_map

        farmer = FourScenFarmer()
        m = create_root(farmer=farmer)
        root_vars = list(m.devoted_acreage.values())
        m.benders = BendersCutGenerator()
        m.benders.set_input(root_vars=root_vars, tol=1e-8)
        for s in farmer.scenarios:
            subproblem_fn_kwargs = dict()
            subproblem_fn_kwargs['root'] = m
            subproblem_fn_kwargs['farmer'] = farmer
            subproblem_fn_kwargs['scenario'] = s
            m.benders.add_subproblem(
                subproblem_fn=create_subproblem,
                subproblem_fn_kwargs=subproblem_fn_kwargs,
                root_eta=m.eta[s],
                subproblem_solver=mip_name,
            )
        opt = pyo.SolverFactory(mip_name)

        for i in range(30):
            res = opt.solve(m, tee=False)
            cuts_added = m.benders.generate_cut()
            if len(cuts_added) == 0:
                break

        self.assertAlmostEqual(m.devoted_acreage['CORN'].value, 100, 7)
        self.assertAlmostEqual(m.devoted_acreage['SUGAR_BEETS'].value, 250, 7)
        self.assertAlmostEqual(m.devoted_acreage['WHEAT'].value, 150, 7)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
from math import copysign

from pyomo.common.collections import ComponentMap
from pyomo.common.config import document_kwargs_from_configdict
from pyomo.common.modeling import unique_component_name
from pyomo.contrib.gdpopt.algorithm_base_class import _GDPoptAlgorithm
from pyomo.contrib.gdpopt.config_options import (
    _add_oa_configs,
    _add_mip_solver_configs,
    _add_nlp_solver_configs,
    _add_tolerance_configs,
)
from pyomo.contrib.gdpopt.create_oa_subproblems import (
    _get_discrete_problem_and_subproblem,
    add_constraint_list,
)
from pyomo.contrib.gdpopt.cut_generation import add_no_good_cut
from pyomo.contrib.gdpopt.oa_algorithm_utils import _OAAlgorithmMixIn
from pyomo.contrib.gdpopt.solve_discrete_problem import solve_MILP_discrete_problem
from pyomo.contrib.gdpopt.util import (
    time_code,
    _add_bigm_constraint_to_transformed_model,
)

from pyomo.core import (
    Block,
    Constraint,
    minimize,
    NonNegativeIntegers,
    NonNegativeReals,
    Objective,
    value,
    Var,
    VarList,
)
from pyomo.core.expr import differentiate
from pyomo.core.expr.visitor import identify_variables
from pyomo.gdp import Disjunct
from pyomo.opt.base import SolverFactory
from pyomo.repn import generate_standard_repn

MAX_SYMBOLIC_DERIV_SIZE = 1000
JacInfo = namedtuple('JacInfo', ['mode', 'vars', 'jac'])


@SolverFactory.register(
    'gdpopt.loa',
    doc="The LOA (logic-based outer approximation) Generalized Disjunctive "
    "Programming (GDP) solver",
)
class GDP_LOA_Solver(_GDPoptAlgorithm, _OAAlgorithmMixIn):
    """The GDPopt (Generalized Disjunctive Programming optimizer) logic-based
    outer approximation (LOA) solver.

    Accepts models that can include nonlinear, continuous variables and
    constraints, as well as logical conditions. For nonconvex problems, LOA
    may not report rigorous dual bounds.
    """

    CONFIG = _GDPoptAlgorithm.CONFIG()
    _add_oa_configs(CONFIG)
    _add_mip_solver_configs(CONFIG)
    _add_nlp_solver_configs(CONFIG, default_solver='ipopt')
    _add_tolerance_configs(CONFIG)

    algorithm = 'LOA'

    # Override solve() to customize the docstring for this solver
    @document_kwargs_from_configdict(CONFIG, doc=_GDPoptAlgorithm.solve.__doc__)
    def solve(self, model, **kwds):
        return super().solve(model, **kwds)

    def _log_citation(self, config):
        config.logger.info("\n" + """- LOA algorithm:
        Türkay, M; Grossmann, IE.
        Logic-based MINLP algorithms for the optimal synthesis of process
        networks. Comp. and Chem. Eng. 1996, 20(8), 959–978.
        DOI: 10.1016/0098-1354(95)00219-7.
        """.strip())

    def _solve_gdp(self, original_model, config):
        logger = config.logger

        # We'll need these to get dual info after solving subproblems
        add_constraint_list(self.original_util_block)

        discrete_problem_util_block, subproblem_util_block = (
            _get_discrete_problem_and_subproblem(self, config)
        )

        discrete = discrete_problem_util_block.parent_block()
        subproblem = subproblem_util_block.parent_block()

        original_obj = self._setup_augmented_penalty_objective(
            discrete_problem_util_block
        )

        self._log_header(logger)

        # main loop
        while not config.iterlim or self.iteration < config.iterlim:
            self.iteration += 1

            # solve linear discrete problem
            with time_code(self.timing, 'mip'):
                oa_obj = self._update_augmented_penalty_objective(
                    discrete_problem_util_block, original_obj, config.OA_penalty_factor
                )
                mip_feasible = solve_MILP_discrete_problem(
                    discrete_problem_util_block, self, config
                )
                self._update_bounds_after_discrete_problem_solve(
                    mip_feasible, oa_obj, logger
                )

            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

            with time_code(self.timing, 'nlp'):
                self._fix_discrete_soln_solve_subproblem_and_add_cuts(
                    discrete_problem_util_block, subproblem_util_block, config
                )

            # Add integer cut
            with time_code(self.timing, "integer cut generation"):
                add_no_good_cut(discrete_problem_util_block, config)

            # Check termination conditions
            if self.any_termination_criterion_met(config):
                break

    def _setup_augmented_penalty_objective(self, discrete_problem_util_block):
        m = discrete_problem_util_block.parent_block()
        discrete_objective = next(m.component_data_objects(Objective, active=True))

        # Set up augmented penalty objective
        discrete_objective.deactivate()
        # placeholder for OA objective
        discrete_problem_util_block.oa_obj = Objective(sense=minimize)

        return discrete_objective

    def _update_augmented_penalty_objective(
        self, discrete_problem_util_block, discrete_objective, OA_penalty_factor
    ):
        m = discrete_problem_util_block.parent_block()
        sign_adjust = 1 if discrete_objective.sense == minimize else -1
        OA_penalty_expr = (
            sign_adjust
            * OA_penalty_factor
            * sum(
                v
                for v in m.component_data_objects(
                    ctype=Var, descend_into=(Block, Disjunct)
                )
                if v.parent_component().local_name == 'GDPopt_OA_slacks'
            )
        )
        discrete_problem_util_block.oa_obj.expr = (
            discrete_objective.expr + OA_penalty_expr
        )

        return discrete_problem_util_block.oa_obj.expr

    def _add_cuts_to_discrete_problem(
        self,
        subproblem_util_block,
        discrete_problem_util_block,
        objective_sense,
        config,
        timing,
    ):
        """Add outer approximation cuts to the linear GDP model."""
        m = discrete_problem_util_block.parent_block()
        nlp = subproblem_util_block.parent_block()
        sign_adjust = -1 if objective_sense == minimize else 1
        # Dictionary mapping blocks to their child blocks we use to store OA
        # cuts. We do it this way because we don't know for sure we can have any
        # given name since we are sticking these on a clone of a user-generated
        # model. But this keeps track that we find one if we've already created
        # it, so that we can add the cuts as indexed constraints.
        if hasattr(discrete_problem_util_block, 'oa_cut_blocks'):
            oa_cut_blocks = discrete_problem_util_block.oa_cut_blocks
        else:
            oa_cut_blocks = discrete_problem_util_block.oa_cut_blocks = dict()

        for discrete_var, subprob_var in zip(
            discrete_problem_util_block.algebraic_variable_list,
            subproblem_util_block.algebraic_variable_list,
        ):
            val = subprob_var.value
            if val is not None and not discrete_var.fixed:
                discrete_var.set_value(val, skip_validation=True)

        config.logger.debug('Adding OA cuts.')

        counter = 0
        if not hasattr(discrete_problem_util_block, 'jacobians'):
            discrete_problem_util_block.jacobians = ComponentMap()
        for constr, subprob_constr in zip(
            discrete_problem_util_block.constraint_list,
            subproblem_util_block.constraint_list,
        ):
            dual_value = nlp.dual.get(subprob_constr, None)
            if dual_value is None or generate_standard_repn(constr.body).is_linear():
                continue

            # Determine if the user pre-specified that OA cuts should not be
            # generated for the given constraint.
            parent_block = constr.parent_block()
            ignore_set = getattr(parent_block, 'GDPopt_ignore_OA', None)
            config.logger.debug('Ignore_set %s' % ignore_set)
            if ignore_set and (
                constr in ignore_set or constr.parent_component() in ignore_set
            ):
                config.logger.debug(
                    'OA cut addition for %s skipped because it is in '
                    'the ignore set.' % constr.name
                )
                continue

            config.logger.debug(
                "Adding OA cut for %s with dual value %s" % (constr.name, dual_value)
            )

            # Cache jacobian
            jacobian = discrete_problem_util_block.jacobians.get(constr, None)
            if jacobian is None:
                constr_vars = list(identify_variables(constr.body, include_fixed=False))
                if len(constr_vars) >= MAX_SYMBOLIC_DERIV_SIZE:
                    mode = differentiate.Modes.reverse_numeric
                else:
                    mode = differentiate.Modes.sympy

                try:
                    jac_list = differentiate(
                        constr.body, wrt_list=constr_vars, mode=mode
                    )
                    jac_map = ComponentMap(zip(constr_vars, jac_list))
                except:
                    if mode is differentiate.Modes.reverse_numeric:
                        raise
                    mode = differentiate.Modes.reverse_numeric
                    jac_map = ComponentMap()
                jacobian = JacInfo(mode=mode, vars=constr_vars, jac=jac_map)
                discrete_problem_util_block.jacobians[constr] = jacobian
            # Recompute numeric derivatives
            if not jacobian.jac:
                jac_list = differentiate(
                    constr.body, wrt_list=jacobian.vars, mode=jacobian.mode
                )
                jacobian.jac.update(zip(jacobian.vars, jac_list))

            # Create a block on which to put outer approximation cuts, if we
            # don't have one already on this parent block.
            oa_utils = oa_cut_blocks.get(parent_block)
            if oa_utils is None:
                nm = unique_component_name(parent_block, 'GDPopt_OA_cuts')
                oa_utils = Block(
                    doc="Block holding outer approximation cuts and associated data."
                )
                parent_block.add_component(nm, oa_utils)
                oa_cut_blocks[parent_block] = oa_utils
                oa_utils.cuts = Constraint(NonNegativeIntegers)
            discrete_prob_oa_utils = discrete_problem_util_block.component(
                'GDPopt_OA_slacks'
            )
            if discrete_prob_oa_utils is None:
                discrete_prob_oa_utils = (
                    discrete_problem_util_block.GDPopt_OA_slacks
                ) = Block(
                    doc="Block holding outer approximation "
                    "slacks for the whole model (so that the "
                    "writers can find them)."
                )
                discrete_prob_oa_utils.slacks = VarList(
                    bounds=(0, config.max_slack), domain=NonNegativeReals, initialize=0
                )

            oa_cuts = oa_utils.cuts
            slack_var = discrete_prob_oa_utils.slacks.add()
            rhs = value(constr.lower) if constr.has_lb() else value(constr.upper)
            try:
                new_oa_cut = (
                    copysign(1, sign_adjust * dual_value)
                    * (
                        value(constr.body)
                        - rhs
                        + sum(
                            value(jac) * (var - value(var))
                            for var, jac in jacobian.jac.items()
                        )
                    )
                    - slack_var
                    <= 0
                )
                assert new_oa_cut.polynomial_degree() in (1, 0)
                idx = len(oa_cuts)
                oa_cuts[idx] = new_oa_cut
                _add_bigm_constraint_to_transformed_model(m, oa_cuts[idx], oa_cuts)
                config.logger.debug("Cut expression: %s" % new_oa_cut)
                counter += 1
            except ZeroDivisionError:
                config.logger.warning(
                    "Zero division occurred attempting to generate OA cut for "
                    "constraint %s.\n"
                    "Skipping OA cut generation for this constraint." % (constr.name,)
                )
                # Simply continue on to the next constraint.
            # Clear out the numeric Jacobian values
            if jacobian.mode is differentiate.Modes.reverse_numeric:
                jacobian.jac.clear()

        config.logger.debug('Added %s OA cuts' % counter)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import operator

from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.shutdown import python_is_shutting_down
from pyomo.core.staleflag import StaleFlagManager
from pyomo.repn.plugins.standard_form import LinearStandardFormCompiler

from pyomo.contrib.solver.common.util import (
    NoDualsError,
    NoReducedCostsError,
    NoSolutionError,
    IncompatibleModelError,
)
from pyomo.contrib.solver.common.solution_loader import SolutionLoaderBase
from .gurobi_direct_base import (
    GurobiDirectBase,
    gurobipy,
    GurobiDirectSolutionLoaderBase,
)
import logging

logger = logging.getLogger(__name__)


class GurobiDirectSolutionLoader(GurobiDirectSolutionLoaderBase):
    def __init__(self, solver_model, pyomo_vars, gurobi_vars, con_map) -> None:
        super().__init__(solver_model)
        self._pyomo_vars = pyomo_vars
        self._gurobi_vars = gurobi_vars
        self._con_map = con_map

    def _get_var_lists(self):
        return self._pyomo_vars, self._gurobi_vars

    def _get_var_map(self):
        return ComponentMap(zip(*self._get_var_lists()))

    def _get_con_map(self):
        return self._con_map

    def __del__(self):
        super().__del__()
        if python_is_shutting_down():
            return
        # Free the associated model
        if self._solver_model is not None:
            self._var_map = None
            self._con_map = None
            # explicitly release the model
            self._solver_model.dispose()
            self._solver_model = None


class GurobiDirect(GurobiDirectBase):
    _minimum_version = (9, 0, 0)

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self._gurobi_vars = None
        self._pyomo_vars = None

    def _pyomo_gurobi_var_iter(self):
        return zip(self._pyomo_vars, self._gurobi_vars)

    def _create_solver_model(self, pyomo_model, config):
        timer = config.timer

        timer.start('compile_model')
        repn = LinearStandardFormCompiler().write(
            pyomo_model, mixed_form=True, set_sense=None
        )
        timer.stop('compile_model')

        if len(repn.objectives) > 1:
            raise IncompatibleModelError(
                f"The {self.__class__.__name__} solver only supports models "
                f"with zero or one objectives (received {len(repn.objectives)})."
            )

        timer.start('prepare_matrices')
        inf = float('inf')
        ninf = -inf
        bounds = list(map(operator.attrgetter('bounds'), repn.columns))
        lb = [ninf if _b is None else _b for _b in map(operator.itemgetter(0), bounds)]
        ub = [inf if _b is None else _b for _b in map(operator.itemgetter(1), bounds)]
        CON = gurobipy.GRB.CONTINUOUS
        BIN = gurobipy.GRB.BINARY
        INT = gurobipy.GRB.INTEGER
        vtype = [
            (
                CON
                if v.is_continuous()
                else BIN if v.is_binary() else INT if v.is_integer() else '?'
            )
            for v in repn.columns
        ]
        sense_type = list('=<>')  # Note: ordering matches 0, 1, -1
        sense = [sense_type[r[1]] for r in repn.rows]
        timer.stop('prepare_matrices')

        gurobi_model = gurobipy.Model(env=self.env())

        timer.start('transfer_model')
        x = gurobi_model.addMVar(
            len(repn.columns),
            lb=lb,
            ub=ub,
            obj=repn.c.todense()[0] if repn.c.shape[0] else 0,
            vtype=vtype,
        )
        A = gurobi_model.addMConstr(repn.A, x, sense, repn.rhs)
        if repn.c.shape[0]:
            gurobi_model.setAttr('ObjCon', repn.c_offset[0])
            gurobi_model.setAttr('ModelSense', int(repn.objectives[0].sense))
        # Note: calling gurobi_model.update() here is not
        # necessary (it will happen as part of optimize()):
        # gurobi_model.update()
        timer.stop('transfer_model')

        self._pyomo_vars = repn.columns
        timer.start('tolist')
        self._gurobi_vars = x.tolist()
        timer.stop('tolist')

        timer.start('create maps')
        timer.start('con map')
        con_map = {}
        for row, gc in zip(repn.rows, A.tolist()):
            pc = row.constraint
            if pc in con_map:
                # range constraint
                con_map[pc] = (con_map[pc], gc)
            else:
                con_map[pc] = gc
        timer.stop('con map')
        timer.stop('create maps')
        solution_loader = GurobiDirectSolutionLoader(
            solver_model=gurobi_model,
            pyomo_vars=self._pyomo_vars,
            gurobi_vars=self._gurobi_vars,
            con_map=con_map,
        )
        has_obj = len(repn.objectives) > 0

        return gurobi_model, solution_loader, has_obj



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import enum
from typing import Optional, Tuple
from datetime import datetime

from pyomo.common.config import (
    ConfigDict,
    ConfigValue,
    IsInstance,
    NonNegativeInt,
    In,
    NonNegativeFloat,
    ADVANCED_OPTION,
    DEVELOPER_OPTION,
)
from pyomo.opt.results.solution import SolutionStatus as LegacySolutionStatus
from pyomo.opt.results.solver import (
    TerminationCondition as LegacyTerminationCondition,
    SolverStatus as LegacySolverStatus,
)


class TerminationCondition(enum.Enum):
    """
    An Enum that enumerates all possible exit statuses for a solver call.

    """

    convergenceCriteriaSatisfied = 0
    "The solver exited because convergence criteria of the problem were satisfied."

    maxTimeLimit = 1
    """The solver exited due to reaching a specified time limit."""

    iterationLimit = 2
    """The solver exited due to reaching a specified iteration limit."""

    objectiveLimit = 3
    """The solver exited due to reaching an objective limit. For example, in
    Gurobi, the exit message "Optimal objective for model was proven to
    be worse than the value specified in the Cutoff parameter" would map
    to objectiveLimit.
    """

    minStepLength = 4
    """The solver exited due to a minimum step length.  Minimum step length
    reached may mean that the problem is infeasible or that the problem
    is feasible but the solver could not converge.
    """

    unbounded = 5
    "The solver exited because the problem has been found to be unbounded."

    provenInfeasible = 6
    "The solver exited because the problem has been proven infeasible."

    locallyInfeasible = 7
    """The solver exited because no feasible solution was found to the
    submitted problem, but it could not be proven that no such solution
    exists.
    """

    infeasibleOrUnbounded = 8
    """Some solvers do not specify between infeasibility or unboundedness
    and instead return that one or the other has occurred. For example,
    in Gurobi, this may occur because there are some steps in presolve
    that prevent Gurobi from distinguishing between infeasibility and
    unboundedness.
    """

    error = 9
    """The solver exited with some error. The error message will also be
    captured and returned.
    """

    interrupted = 10
    "The solver was interrupted while running."

    licensingProblems = 11
    """The solver experienced issues with licensing. This could be that no
    license was found, the license is of the wrong type for the problem
    (e.g., problem is too big for type of license), or there was an
    issue contacting a licensing server.
    """

    emptyModel = 12
    "The model being solved did not have any variables"

    unknown = 42
    "All other unrecognized exit statuses fall in this category."


class SolutionStatus(enum.Enum):
    """An enumeration for interpreting the result of a termination. This
    describes the designated status by the solver to be loaded back into
    the model.

    """

    noSolution = 0
    """No (single) solution was found; possible that a population of
    solutions was returned.
    """

    unknown = 5
    "Solution returned, but feasibility/optimality unknown."

    infeasible = 10
    "Solution point does not satisfy some domains and/or constraints."

    feasible = 20
    "A solution for which all of the constraints in the model are satisfied."

    optimal = 30
    "A feasible solution satisfying the solver's optimality criteria."


class Results(ConfigDict):
    """Base class for all solver results"""

    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )

        self.solution_loader = self.declare(
            'solution_loader',
            ConfigValue(
                description="Object for loading the solution back into the model.",
                visibility=DEVELOPER_OPTION,
            ),
        )
        self.termination_condition: TerminationCondition = self.declare(
            'termination_condition',
            ConfigValue(
                domain=In(TerminationCondition),
                default=TerminationCondition.unknown,
                description="The reason the solver exited. This is a member of the "
                "TerminationCondition enum.",
            ),
        )
        self.solution_status: SolutionStatus = self.declare(
            'solution_status',
            ConfigValue(
                domain=In(SolutionStatus),
                default=SolutionStatus.noSolution,
                description="The result of the solve call. This is a member of "
                "the SolutionStatus enum.",
            ),
        )
        self.incumbent_objective: Optional[float] = self.declare(
            'incumbent_objective',
            ConfigValue(
                domain=float,
                default=None,
                description="If a feasible solution was found, this is the objective "
                "value of the best solution found. If no feasible solution was found, this is None.",
            ),
        )
        self.objective_bound: Optional[float] = self.declare(
            'objective_bound',
            ConfigValue(
                domain=float,
                default=None,
                description="The best objective bound found. For minimization problems, "
                "this is the lower bound. For maximization problems, this is the "
                "upper bound. For solvers that do not provide an objective bound, "
                "this should be -inf (minimization) or inf (maximization)",
            ),
        )
        self.solver_name: Optional[str] = self.declare(
            'solver_name',
            ConfigValue(domain=str, description="The name of the solver in use."),
        )
        self.solver_version: Optional[Tuple[int, ...]] = self.declare(
            'solver_version',
            ConfigValue(
                domain=tuple,
                description="A tuple representing the version of the solver in use.",
            ),
        )
        self.timing_info: ConfigDict = self.declare(
            'timing_info', ConfigDict(implicit=True)
        )

        self.timing_info.start_timestamp: datetime = self.timing_info.declare(
            'start_timestamp',
            ConfigValue(
                domain=IsInstance(datetime),
                description="UTC timestamp of when run was initiated.",
            ),
        )
        self.timing_info.wall_time: Optional[float] = self.timing_info.declare(
            'wall_time',
            ConfigValue(
                domain=NonNegativeFloat,
                description="Elapsed wall clock time for entire process.",
            ),
        )
        self.extra_info: ConfigDict = self.declare(
            'extra_info', ConfigDict(implicit=True)
        )
        self.solver_config: ConfigDict = self.declare(
            'solver_config',
            ConfigValue(
                description="A copy of the config object used in the solve call.",
                visibility=ADVANCED_OPTION,
            ),
        )
        self.solver_log: str = self.declare(
            'solver_log',
            ConfigValue(
                domain=str,
                default=None,
                visibility=ADVANCED_OPTION,
                description="Any solver log messages.",
            ),
        )

    def display(
        self, content_filter=None, indent_spacing=2, ostream=None, visibility=0
    ):
        return super().display(content_filter, indent_spacing, ostream, visibility)


# Everything below here preserves backwards compatibility

legacy_termination_condition_map = {
    TerminationCondition.unknown: LegacyTerminationCondition.unknown,
    TerminationCondition.maxTimeLimit: LegacyTerminationCondition.maxTimeLimit,
    TerminationCondition.iterationLimit: LegacyTerminationCondition.maxIterations,
    TerminationCondition.objectiveLimit: LegacyTerminationCondition.minFunctionValue,
    TerminationCondition.minStepLength: LegacyTerminationCondition.minStepLength,
    TerminationCondition.convergenceCriteriaSatisfied: LegacyTerminationCondition.optimal,
    TerminationCondition.unbounded: LegacyTerminationCondition.unbounded,
    TerminationCondition.provenInfeasible: LegacyTerminationCondition.infeasible,
    TerminationCondition.locallyInfeasible: LegacyTerminationCondition.infeasible,
    TerminationCondition.infeasibleOrUnbounded: LegacyTerminationCondition.infeasibleOrUnbounded,
    TerminationCondition.error: LegacyTerminationCondition.error,
    TerminationCondition.interrupted: LegacyTerminationCondition.resourceInterrupt,
    TerminationCondition.licensingProblems: LegacyTerminationCondition.licensingProblems,
}


legacy_solver_status_map = {
    TerminationCondition.unknown: LegacySolverStatus.unknown,
    TerminationCondition.maxTimeLimit: LegacySolverStatus.aborted,
    TerminationCondition.iterationLimit: LegacySolverStatus.aborted,
    TerminationCondition.objectiveLimit: LegacySolverStatus.aborted,
    TerminationCondition.minStepLength: LegacySolverStatus.error,
    TerminationCondition.convergenceCriteriaSatisfied: LegacySolverStatus.ok,
    TerminationCondition.unbounded: LegacySolverStatus.error,
    TerminationCondition.provenInfeasible: LegacySolverStatus.error,
    TerminationCondition.locallyInfeasible: LegacySolverStatus.error,
    TerminationCondition.infeasibleOrUnbounded: LegacySolverStatus.error,
    TerminationCondition.error: LegacySolverStatus.error,
    TerminationCondition.interrupted: LegacySolverStatus.aborted,
    TerminationCondition.licensingProblems: LegacySolverStatus.error,
}


# The logic for the new solution status to legacy solution status
# cannot be contained in a dictionary because it is a many -> one
# relationship.
def legacy_solution_status_map(results):
    """
    Map the new TerminationCondition/SolutionStatus values into LegacySolutionStatus
    objects. Because we condensed results objects, some of the previous statuses
    are no longer clearly achievable.
    """
    if results.termination_condition in set(
        [
            TerminationCondition.maxTimeLimit,
            TerminationCondition.iterationLimit,
            TerminationCondition.objectiveLimit,
            TerminationCondition.minStepLength,
        ]
    ):
        return LegacySolutionStatus.stoppedByLimit
    if results.termination_condition in set(
        [TerminationCondition.provenInfeasible, TerminationCondition.locallyInfeasible]
    ):
        return LegacySolutionStatus.infeasible
    if results.termination_condition in set(
        [
            TerminationCondition.error,
            TerminationCondition.licensingProblems,
            TerminationCondition.interrupted,
        ]
    ):
        return LegacySolutionStatus.error
    if results.termination_condition == TerminationCondition.unbounded:
        return LegacySolutionStatus.unbounded
    if (
        results.termination_condition
        == TerminationCondition.convergenceCriteriaSatisfied
    ):
        if results.solution_status == SolutionStatus.feasible:
            return LegacySolutionStatus.feasible
        return LegacySolutionStatus.optimal
    return LegacySolutionStatus.unknown



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Provides functions for retrieving disjunctive variable bound information stored on a model."""

from pyomo.common.collections import ComponentMap
from pyomo.core import value

inf = float('inf')


def disjunctive_bounds(scope):
    """Return all of the variable bounds defined at a disjunctive scope."""
    possible_disjunct = scope
    while possible_disjunct is not None:
        try:
            return possible_disjunct._disj_var_bounds
        except AttributeError:
            # possible disjunct does not have attribute '_disj_var_bounds'.
            # Try again with the scope's parent block.
            possible_disjunct = possible_disjunct.parent_block()
    # Unable to find '_disj_var_bounds' attribute within search scope.
    return ComponentMap()


def disjunctive_bound(var, scope):
    """Compute the disjunctive bounds for a variable in a given scope.

    Args:
        var (VarData): Variable for which to compute bound
        scope (Component): The scope in which to compute the bound. If not a
            DisjunctData, it will walk up the tree and use the scope of the
            most immediate enclosing DisjunctData.

    Returns:
        numeric: the tighter of either the disjunctive lower bound, the
            variable lower bound, or (-inf, inf) if neither exist.

    """
    # Initialize to the global variable bound
    var_bnd = (
        value(var.lb) if var.has_lb() else -inf,
        value(var.ub) if var.has_ub() else inf,
    )
    possible_disjunct = scope
    while possible_disjunct is not None:
        try:
            disj_bnd = possible_disjunct._disj_var_bounds.get(var, (-inf, inf))
            disj_bnd = (max(var_bnd[0], disj_bnd[0]), min(var_bnd[1], disj_bnd[1]))
            return disj_bnd
        except AttributeError:
            # possible disjunct does not have attribute '_disj_var_bounds'.
            # Try again with the scope's parent block.
            possible_disjunct = possible_disjunct.parent_block()
    # Unable to find '_disj_var_bounds' attribute within search scope.
    return var_bnd


def disjunctive_lb(var, scope):
    """Compute the disjunctive lower bound for a variable in a given scope."""
    return disjunctive_bound(var, scope)[0]


def disjunctive_ub(var, scope):
    """Compute the disjunctive upper bound for a variable in a given scope."""
    return disjunctive_bound(var, scope)[1]



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# -*- coding: utf-8 -*-
"""Transformation to fix and enforce disjunct True/False status."""

import logging
from math import fabs

from pyomo.common.config import ConfigDict, ConfigValue, NonNegativeFloat
from pyomo.contrib.cp.transform.logical_to_disjunctive_program import (
    LogicalToDisjunctive,
)
from pyomo.core.base import Transformation, TransformationFactory
from pyomo.core.base.block import Block
from pyomo.core.expr.numvalue import value
from pyomo.gdp import GDP_Error
from pyomo.gdp.disjunct import Disjunct, Disjunction
from pyomo.gdp.plugins.bigm import BigM_Transformation

logger = logging.getLogger('pyomo.gdp.fix_disjuncts')


def _transformation_name_or_object(transformation_name_or_object):
    if isinstance(transformation_name_or_object, Transformation):
        return transformation_name_or_object
    xform = TransformationFactory(transformation_name_or_object)
    if xform is None:
        raise ValueError(
            "Expected valid name for a registered Pyomo transformation. "
            "\n\tRecieved: %s" % transformation_name_or_object
        )
    return xform


@TransformationFactory.register(
    'gdp.fix_disjuncts',
    doc="""Fix disjuncts to their current Boolean values and transforms any
    LogicalConstraints and BooleanVars so that the resulting model is a
    (MI)(N)LP.""",
)
class GDP_Disjunct_Fixer(Transformation):
    """Fix disjuncts to their current Boolean values.

    This reclassifies all disjuncts in the passed model instance as ctype Block
    and deactivates the constraints and disjunctions within inactive disjuncts.
    In addition, it transforms relevant LogicalConstraints and BooleanVars so
    that the resulting model is a (MI)(N)LP (where it is only mixed-integer
    if the model contains integer-domain Vars or BooleanVars which were not
    indicator_vars of Disjuncs.
    """

    def __init__(self, **kwargs):
        # TODO This uses an integer tolerance. At some point, these should be
        # standardized.
        super(GDP_Disjunct_Fixer, self).__init__(**kwargs)

    CONFIG = ConfigDict("gdp.fix_disjuncts")
    CONFIG.declare(
        "GDP_to_MIP_transformation",
        ConfigValue(
            default=BigM_Transformation(),
            domain=_transformation_name_or_object,
            description="The name of the transformation to call after the "
            "'logical_to_disjunctive' transformation in order to finish "
            "transforming to a MI(N)LP.",
            doc="""
        If there are no logical constraints on the model being transformed,
        this option is not used. However, if there are logical constraints
        that involve mixtures of Boolean and integer variables, this option
        specifies the transformation to use to transform the model with fixed
        Disjuncts to a MI(N)LP. Uses 'gdp.bigm' by default.
        """,
        ),
    )

    def _apply_to(self, model, **kwds):
        """Fix all disjuncts in the given model and reclassify them to
        Blocks."""
        config = self.config = self.CONFIG(kwds.pop('options', {}))
        config.set_value(kwds)

        self._transformContainer(model)

        # Reclassify all disjuncts
        for disjunct_object in model.component_objects(
            Disjunct, descend_into=(Block, Disjunct)
        ):
            disjunct_object.parent_block().reclassify_component_type(
                disjunct_object, Block
            )

        # Transform any remaining logical stuff
        TransformationFactory('contrib.logical_to_disjunctive').apply_to(model)
        # Transform anything disjunctive that the above created:
        config.GDP_to_MIP_transformation.apply_to(model)

    def _transformContainer(self, obj):
        """Find all disjuncts in the container and transform them."""
        for disjunct in obj.component_data_objects(
            ctype=Disjunct, active=True, descend_into=True
        ):
            _bool = disjunct.indicator_var
            if _bool.value is None:
                raise GDP_Error(
                    "The value of the indicator_var of "
                    "Disjunct '%s' is None. All indicator_vars "
                    "must have values before calling "
                    "'fix_disjuncts'." % disjunct.name
                )
            elif _bool.value:
                disjunct.indicator_var.fix(True)
                self._transformContainer(disjunct)
            else:
                # Deactivating fixes the indicator_var to False
                disjunct.deactivate()

        for disjunction in obj.component_data_objects(
            ctype=Disjunction, active=True, descend_into=True
        ):
            self._transformDisjunctionData(disjunction)

    def _transformDisjunctionData(self, disjunction):
        # the sum of all the indicator variable values of disjuncts in the
        # disjunction
        logical_sum = sum(
            value(disj.binary_indicator_var) for disj in disjunction.disjuncts
        )

        # Check that the disjunctions are not being fixed to an infeasible
        # realization.
        if disjunction.xor and not logical_sum == 1:
            # for XOR disjunctions, the sum of all disjunct values should be 1
            raise GDP_Error(
                "Disjunction %s violated. "
                "Expected 1 disjunct to be active, but %s were active."
                % (disjunction.name, logical_sum)
            )
        elif not logical_sum >= 1:
            # for non-XOR disjunctions, the sum of all disjunct values should
            # be at least 1
            raise GDP_Error(
                "Disjunction %s violated. "
                "Expected at least 1 disjunct to be active, "
                "but none were active."
            )
        else:
            # disjunction is in feasible realization. Deactivate it.
            disjunction.deactivate()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

#
# Class for reading an AMPL *.sol file
#

import re

from pyomo.opt.base import results
from pyomo.opt.base.formats import ResultsFormat
from pyomo.opt import SolverResults, SolutionStatus, SolverStatus, TerminationCondition


@results.ReaderFactory.register(str(ResultsFormat.sol))
class ResultsReader_sol(results.AbstractResultsReader):
    """
    Class that reads in a ``*.sol`` results file and generates a
    SolverResults object.
    """

    def __init__(self, name=None):
        results.AbstractResultsReader.__init__(self, ResultsFormat.sol)
        if not name is None:
            self.name = name

    def __call__(self, filename, res=None, soln=None, suffixes=[]):
        """
        Parse a ``*.sol`` file
        """
        try:
            with open(filename, "r") as f:
                return self._load(f, res, soln, suffixes)
        except ValueError as e:
            with open(filename, "r") as f:
                fdata = f.read()
            raise ValueError(
                "Error reading '%s': %s.\n"
                "SOL File Output:\n%s" % (filename, str(e), fdata)
            )

    def _load(self, fin, res, soln, suffixes):
        if res is None:
            res = SolverResults()
        #
        # Some solvers (minto) do not write a message.  We will assume
        # all non-blank lines up the 'Options' line is the message.
        msg = []
        while True:
            line = fin.readline()
            if not line:
                # EOF
                break
            line = line.strip()
            if line == 'Options':
                break
            if line:
                msg.append(line)
        msg = '\n'.join(msg)
        z = []
        if line[:7] == "Options":
            line = fin.readline()
            nopts = int(line)
            need_vbtol = False
            if nopts > 4:  # WEH - when is this true?
                nopts -= 2
                need_vbtol = True
            for i in range(nopts + 4):
                line = fin.readline()
                z += [int(line)]
            if need_vbtol:  # WEH - when is this true?
                line = fin.readline()
                z += [float(line)]
        else:
            raise ValueError("no Options line found")
        n = z[nopts + 3]  # variables
        m = z[nopts + 1]  # constraints
        x = []
        y = []
        i = 0
        while i < m:
            line = fin.readline()
            y.append(float(line))
            i += 1
        i = 0
        while i < n:
            line = fin.readline()
            x.append(float(line))
            i += 1
        objno = [0, 0]
        line = fin.readline()
        if line:  # WEH - when is this true?
            if line[:5] != "objno":  # pragma:nocover
                raise ValueError("expected 'objno', found '%s'" % (line))
            t = line.split()
            if len(t) != 3:
                raise ValueError(
                    "expected two numbers in objno line, but found '%s'" % (line)
                )
            objno = [int(t[1]), int(t[2])]
        res.solver.message = msg.strip()
        res.solver.message = res.solver.message.replace("\n", "; ")
        if isinstance(res.solver.message, str):
            res.solver.message = res.solver.message.replace(':', '\\x3a')
        ##res.solver.instanceName = osrl.header.instanceName
        ##res.solver.systime = osrl.header.time
        res.solver.status = SolverStatus.ok
        soln_status = SolutionStatus.unknown
        if (objno[1] >= 0) and (objno[1] <= 99):
            objno_message = "OPTIMAL SOLUTION FOUND!"
            res.solver.termination_condition = TerminationCondition.optimal
            res.solver.status = SolverStatus.ok
            soln_status = SolutionStatus.optimal
        elif (objno[1] >= 100) and (objno[1] <= 199):
            objno_message = "Optimal solution indicated, but ERROR LIKELY!"
            res.solver.termination_condition = TerminationCondition.optimal
            res.solver.status = SolverStatus.warning
            soln_status = SolutionStatus.optimal
        elif (objno[1] >= 200) and (objno[1] <= 299):
            objno_message = "INFEASIBLE SOLUTION: constraints cannot be satisfied!"
            res.solver.termination_condition = TerminationCondition.infeasible
            res.solver.status = SolverStatus.warning
            soln_status = SolutionStatus.infeasible
        elif (objno[1] >= 300) and (objno[1] <= 399):
            objno_message = (
                "UNBOUNDED PROBLEM: the objective can be improved without limit!"
            )
            res.solver.termination_condition = TerminationCondition.unbounded
            res.solver.status = SolverStatus.warning
            soln_status = SolutionStatus.unbounded
        elif (objno[1] >= 400) and (objno[1] <= 499):
            objno_message = (
                "EXCEEDED MAXIMUM NUMBER OF ITERATIONS: the solver "
                "was stopped by a limit that you set!"
            )
            res.solver.termination_condition = TerminationCondition.maxIterations
            res.solver.status = SolverStatus.warning
            soln_status = SolutionStatus.stoppedByLimit
        elif (objno[1] >= 500) and (objno[1] <= 599):
            objno_message = (
                "FAILURE: the solver stopped by an error condition "
                "in the solver routines!"
            )
            res.solver.termination_condition = TerminationCondition.internalSolverError
            res.solver.status = SolverStatus.error
            soln_status = SolutionStatus.error
        res.solver.id = objno[1]
        ##res.problem.name = osrl.header.instanceName
        if res.solver.termination_condition in [
            TerminationCondition.unknown,
            TerminationCondition.maxIterations,
            TerminationCondition.minFunctionValue,
            TerminationCondition.minStepLength,
            TerminationCondition.globallyOptimal,
            TerminationCondition.locallyOptimal,
            TerminationCondition.optimal,
            TerminationCondition.maxEvaluations,
            TerminationCondition.other,
            TerminationCondition.infeasible,
        ]:
            if soln is None:
                soln = res.solution.add()
            res.solution.status = soln_status
            soln.status_description = objno_message
            soln.message = msg.strip()
            soln.message = res.solver.message.replace("\n", "; ")
            soln_variable = soln.variable
            i = 0
            for var_value in x:
                soln_variable["v" + str(i)] = {"Value": var_value}
                i = i + 1
            soln_constraint = soln.constraint
            if any(re.match(suf, "dual") for suf in suffixes):
                for i in range(0, len(y)):
                    soln_constraint["c" + str(i)] = {"Dual": y[i]}

            ### Read suffixes ###
            line = fin.readline()
            while line:
                line = line.strip()
                if line == "":
                    continue
                line = line.split()
                if line[0] != 'suffix':
                    # We assume this is the start of a
                    # section like kestrel_option, which
                    # comes after all suffixes.
                    remaining = ""
                    line = fin.readline()
                    while line:
                        remaining += line.strip() + "; "
                        line = fin.readline()
                    res.solver.message += remaining
                    break
                unmasked_kind = int(line[1])
                kind = unmasked_kind & 3  # 0-var, 1-con, 2-obj, 3-prob
                convert_function = int
                if (unmasked_kind & 4) == 4:
                    convert_function = float
                nvalues = int(line[2])
                # namelen = int(line[3])
                # tablen = int(line[4])
                tabline = int(line[5])
                suffix_name = fin.readline().strip()
                if any(re.match(suf, suffix_name) for suf in suffixes):
                    # ignore translation of the table number to string value for now,
                    # this information can be obtained from the solver documentation
                    for n in range(tabline):
                        fin.readline()
                    if kind == 0:  # Var
                        for cnt in range(nvalues):
                            suf_line = fin.readline().split()
                            key = "v" + suf_line[0]
                            if key not in soln_variable:
                                soln_variable[key] = {}
                            soln_variable[key][suffix_name] = convert_function(
                                suf_line[1]
                            )
                    elif kind == 1:  # Con
                        for cnt in range(nvalues):
                            suf_line = fin.readline().split()
                            key = "c" + suf_line[0]
                            if key not in soln_constraint:
                                soln_constraint[key] = {}
                            # GH: About the comment below: This makes for a
                            # confusing results object and more confusing tests.
                            # We should not muck with the names of suffixes
                            # coming out of the sol file.
                            #
                            #   convert the first letter of the suffix name to upper case,
                            #   mainly for pretty-print / output purposes. these are lower-cased
                            #   when loaded into real suffixes, so it is largely redundant.
                            translated_suffix_name = (
                                suffix_name[0].upper() + suffix_name[1:]
                            )
                            soln_constraint[key][translated_suffix_name] = (
                                convert_function(suf_line[1])
                            )
                    elif kind == 2:  # Obj
                        for cnt in range(nvalues):
                            suf_line = fin.readline().split()
                            soln.objective.setdefault("o" + suf_line[0], {})[
                                suffix_name
                            ] = convert_function(suf_line[1])
                    elif kind == 3:  # Prob
                        # Skip problem kind suffixes for now. Not sure the
                        # best place to put them in the results object
                        for cnt in range(nvalues):
                            suf_line = fin.readline().split()
                            soln.problem[suffix_name] = convert_function(suf_line[1])
                else:
                    # do not store the suffix in the solution object
                    for cnt in range(nvalues):
                        fin.readline()
                line = fin.readline()

        #
        # This is a bit of a hack to accommodate PICO.  If
        # the PICO parser has parsed the # of constraints, then
        # don't try to read it in from the *.sol file.  The reason
        # is that these may be inconsistent values!
        #
        if res.problem.number_of_constraints == 0:
            res.problem.number_of_constraints = m
        res.problem.number_of_variables = n
        res.problem.number_of_objectives = 1
        return res



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.enums import ObjectiveSense, minimize, maximize
from pyomo.core.expr.numvalue import as_numeric
from pyomo.core.kernel.base import _abstract_readwrite_property
from pyomo.core.kernel.container_utils import define_simple_containers
from pyomo.core.kernel.expression import IExpression


class IObjective(IExpression):
    """
    The interface for optimization objectives.
    """

    __slots__ = ()

    #
    # Implementations can choose to define these
    # properties as using __slots__, __dict__, or
    # by overriding the @property method
    #

    sense = _abstract_readwrite_property(
        doc=("The optimization direction for the objective (minimize or maximize)")
    )

    #
    # Interface
    #

    def is_minimizing(self):
        return self.sense == minimize


class objective(IObjective):
    """An optimization objective."""

    _ctype = IObjective
    __slots__ = ("_parent", "_storage_key", "_active", "_expr", "_sense", "__weakref__")

    def __init__(self, expr=None, sense=minimize):
        self._parent = None
        self._storage_key = None
        self._active = True
        self._expr = None
        self._sense = None

        # call the setters
        self.sense = sense
        self.expr = expr

    #
    # Define the IExpression abstract methods
    #

    @property
    def expr(self):
        return self._expr

    @expr.setter
    def expr(self, expr):
        self._expr = as_numeric(expr) if (expr is not None) else None

    #
    # Define the IObjective abstract methods
    #

    @property
    def sense(self):
        return self._sense

    @sense.setter
    def sense(self, sense):
        """Set the sense (direction) of this objective."""
        self._sense = ObjectiveSense(sense)


# inserts class definitions for simple _tuple, _list, and
# _dict containers into this module
define_simple_containers(globals(), "objective", IObjective)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.dependencies import numpy as np
import pyomo.common.dependencies.scipy.sparse as scipy_sparse
from pyomo.common.dependencies import attempt_import

egb, egb_available = attempt_import(
    'pyomo.contrib.pynumero.interfaces.external_grey_box'
)

if egb_available:

    class GreyBoxModel(egb.ExternalGreyBoxModel):
        """Greybox model to compute the example objective function."""

        def __init__(self, initial, use_exact_derivatives=True, verbose=False):
            """
            Parameters

            use_exact_derivatives: bool
                If True, the exact derivatives are used. If False, the finite difference
                approximation is used.
            verbose: bool
                If True, print information about the model.
            """
            self._use_exact_derivatives = use_exact_derivatives
            self.verbose = verbose
            self.initial = initial

            # For use with exact Hessian
            self._output_con_mult_values = np.zeros(1)

            if not use_exact_derivatives:
                raise NotImplementedError(
                    "use_exact_derivatives == False not supported"
                )

        def input_names(self):
            """Return the names of the inputs."""
            self.input_name_list = ["X1", "X2", "Y1", "Y2", "Y3"]

            return self.input_name_list

        def equality_constraint_names(self):
            """Return the names of the equality constraints."""
            # no equality constraints
            return []

        def output_names(self):
            """Return the names of the outputs."""
            return ['z']

        def set_output_constraint_multipliers(self, output_con_multiplier_values):
            """Set the values of the output constraint multipliers."""
            # because we only have one output constraint
            assert len(output_con_multiplier_values) == 1
            np.copyto(self._output_con_mult_values, output_con_multiplier_values)

        def finalize_block_construction(self, pyomo_block):
            """Finalize the construction of the ExternalGreyBoxBlock."""
            if self.initial is not None:
                print("initialized")
                pyomo_block.inputs["X1"].value = self.initial["X1"]
                pyomo_block.inputs["X2"].value = self.initial["X2"]
                pyomo_block.inputs["Y1"].value = self.initial["Y1"]
                pyomo_block.inputs["Y2"].value = self.initial["Y2"]
                pyomo_block.inputs["Y3"].value = self.initial["Y3"]

            else:
                print("uninitialized")
                for n in self.input_name_list:
                    pyomo_block.inputs[n].value = 1

            pyomo_block.inputs["X1"].setub(4)
            pyomo_block.inputs["X1"].setlb(0)

            pyomo_block.inputs["X2"].setub(4)
            pyomo_block.inputs["X2"].setlb(0)

            pyomo_block.inputs["Y1"].setub(1)
            pyomo_block.inputs["Y1"].setlb(0)

            pyomo_block.inputs["Y2"].setub(1)
            pyomo_block.inputs["Y2"].setlb(0)

            pyomo_block.inputs["Y3"].setub(1)
            pyomo_block.inputs["Y3"].setlb(0)

        def set_input_values(self, input_values):
            """Set the values of the inputs."""
            self._input_values = list(input_values)

        def evaluate_equality_constraints(self):
            """Evaluate the equality constraints."""
            return None

        def evaluate_outputs(self):
            """Evaluate the output of the model."""
            # form matrix as a list of lists
            # M = self._extract_and_assemble_fim()
            x1 = self._input_values[0]
            x2 = self._input_values[1]
            y1 = self._input_values[2]
            y2 = self._input_values[3]
            y3 = self._input_values[4]
            # z
            z = x1**2 + x2**2 + y1 + 1.5 * y2 + 0.5 * y3

            if self.verbose:
                print("\n Consider inputs [x1,x2,y1,y2,y3] =\n", x1, x2, y1, y2, y3)
                print("   z = ", z, "\n")

            return np.asarray([z], dtype=np.float64)

        def evaluate_jacobian_equality_constraints(self):
            """Evaluate the Jacobian of the equality constraints."""
            return None

        """
        def _extract_and_assemble_fim(self):
            M = np.zeros((self.n_parameters, self.n_parameters))
            for i in range(self.n_parameters):
                for k in range(self.n_parameters):                
                    M[i,k] = self._input_values[self.ele_to_order[(i,k)]]

            return M
        """

        def evaluate_jacobian_outputs(self):
            """Evaluate the Jacobian of the outputs."""
            if self._use_exact_derivatives:
                # compute gradient of log determinant
                row = np.zeros(5)  # to store row index
                col = np.zeros(5)  # to store column index
                data = np.zeros(5)  # to store data

                row[0], col[0], data[0] = (0, 0, 2 * self._input_values[0])  # x1
                row[0], col[1], data[1] = (0, 1, 2 * self._input_values[1])  # x2
                row[0], col[2], data[2] = (0, 2, 1)  # y1
                row[0], col[3], data[3] = (0, 3, 1.5)  # y2
                row[0], col[4], data[4] = (0, 4, 0.5)  # y3

                # sparse matrix
                return scipy_sparse.coo_matrix((data, (row, col)), shape=(1, 5))

    def build_model_external(m):
        ex_model = GreyBoxModel(initial={"X1": 0, "X2": 0, "Y1": 0, "Y2": 1, "Y3": 1})
        m.egb = egb.ExternalGreyBoxBlock()
        m.egb.set_external_model(ex_model)

else:
    GreyBoxModel = None
    build_model_external = None
